
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <time.h>
#include <hce_DLNAWiFiInterface.h>
#include <semaphore.h>
#include <log.h>
#include <lnxipc.h>
#include <clients.h>
#include <upnpglobalvars.h>
#include "DualosIpc.h"

#define IPC_RECEIVED_MAX_BUFFER_SIZE 512
#define DLNA_UTILITY_SEMAPHORE_WAIT_FOREVER 0xFFFFFFFF


typedef struct _DLNAWiFiInterfaceInfoStruct_
{
  sem_t   sIPCSemaphore;
  sem_t*  psIPCSemaphorePtr;
  void*   pvServerFD;
  void*   pvClientFD;
  void*   pvIPCFD;
  char    bDeinitFlag;
  char    pbIPCReceivedBuffer[IPC_RECEIVED_MAX_BUFFER_SIZE];
}DLNAWiFiInterfaceInfoStruct;  

#define DLNA_WIFI_INTERFACE_IPC_SEMA_NAME    "Ability DLNA WIFI interface semaphore"
#define DLNA_WIFI_INTERFACE_IPC_SEMA_INICNTR  0
static DLNAWiFiInterfaceInfoStruct* pDLNAWiFiInterfaceInfo = NULL;

typedef enum {
   OS_THREADX = 0,
   OS_LINUX =1,
}os_type;

//-----------------------------------------------------------------------------------------------------
static int Convert_FilePath( char* target , char* source , os_type target_ostype )
{

    char* p = NULL;
    
    if(source == NULL || target == NULL)
    {
        DLNA_Printf(E_DEBUG, L_GENERAL,"NULL %p %p \n", target,source);
        return -1;
    }

    strcpy(target, source);

    if(target_ostype == OS_LINUX)
    {
        p = strchr( source, '\\');
        while(p){
            target[p-source] = '/'; 
            p = strchr( p+1, '\\');
        }
        
    }else if(target_ostype == OS_THREADX)
    {
        p = strchr( source, '/');
        while(p){
            target[p-source] = '\\';
            p = strchr( p+1, '/');
        }
        
    }else{
        DLNA_Printf(E_DEBUG, L_GENERAL,"unknown ostype: %d \n", target_ostype);
        return -1;
    }

    return 0;

}
//-----------------------------------------------------------------------------------------------------
void HceDLNA_IPCServerReceiveCB(void *rcv_buff, void* feedback_buff)
{
    //--------------------Transfer data---------------------------------------
    LNXIPC_GROUP_A_MSG_T* psReqIPC      = (LNXIPC_GROUP_A_MSG_T*)rcv_buff;
    LNXIPC_GROUP_A_MSG_T* psResponseIPC = (LNXIPC_GROUP_A_MSG_T*)feedback_buff;
    DLNAIPCContentReqStruct* psReqInfo    = NULL;
    DLNAIPCContentResStruct* psResponseInfo = NULL;
    void* pvReqData      = NULL;
    void* pvResponseData = NULL;
    if(psReqIPC != NULL)  {
        psReqInfo = (DLNAIPCContentReqStruct*) psReqIPC->content;
        pvReqData = &psReqIPC->content[sizeof(DLNAIPCContentReqStruct)];
    }   
    if(psResponseIPC != NULL)  {        
        psResponseInfo   = (DLNAIPCContentResStruct*) psResponseIPC->content;
        pvResponseData = &psResponseIPC->content[sizeof(DLNAIPCContentResStruct)];
    }    
    psResponseInfo->eErrorResult = DLNA_REMOTE_MSG_RESULT_TYPE_OK;
    //-----------------------------------------------------------------------
    switch(psReqIPC->msgid) {

        case HCE_WIFI_MSGID_APP_SCAN:
        {
            int i;
            int used_size=0;
            int rarry[20];
            int number = 20; // max 20 result
            struct client_cache_s* pclient;
            char* p;
            HCE_DLNA_MSGID_APP_SCAN_RES_PARA ScanAPPList = {0};
            /* send M-SEARCH here? currently server will send M-SEARCH per 5 seconds */


            /* stop sending M-SEARCH, keep the last list we are publishing */

            /* get list */
            GetClientCachesByModelName("LUMIX_UP_SERVER_SMP", rarry, &number);
            p = (char*)(ScanAPPList.servers);
            if(number > 0){
                
                for(i=0; i<number; i++){
                    pclient = &(clients[rarry[i]]);
                    
                    if(pclient->friendlyname != NULL){
                        used_size = used_size + strlen(pclient->friendlyname)+1;
                        if(used_size > sizeof(ScanAPPList.servers))
                            break; /* out of bundery, stop copy */
                        strcpy(p, pclient->friendlyname);
                        DLNA_Printf(E_DEBUG, L_GENERAL,"[%d] %s \n", rarry[i], p);
                        p += used_size;
                        
                    }else{
                    /* if no friendlyname, copy " " */
                        used_size = used_size + strlen(" ")+1;
                        if(used_size > sizeof(ScanAPPList.servers))
                            break; /* out of bundery, stop copy */
                        strcpy(p, "DeviceX");
                        DLNA_Printf(E_DEBUG, L_GENERAL,"[%d] %s \n", rarry[i], p);
                        p += used_size;
                    }

                    ScanAPPList.idxs[i] = rarry[i];
                    ScanAPPList.number++;
                }

            }else{
                ScanAPPList.number = 0;
            }
            
            memcpy(pvResponseData, &ScanAPPList, sizeof(HCE_DLNA_MSGID_APP_SCAN_RES_PARA));
        }
        break;
        case HCE_WIFI_MSGID_APP_SELECT:
        {
            DLNA_Printf(E_DEBUG, L_GENERAL,"Recevied MSG HCE_WIFI_MSGID_APP_SELECT\n");
            HCE_DLNA_MSGID_SELECT_SERVER_REQ_PARA* SelectedServer =
                (HCE_DLNA_MSGID_SELECT_SERVER_REQ_PARA*) pvReqData;
            DLNA_Printf(E_DEBUG, L_GENERAL,"Server Name: %s, Server Index = %d\n", SelectedServer->serverName, SelectedServer->serverIdx);
            SaveSelectServer(SelectedServer->serverName, &(SelectedServer->serverIdx));
        }
        break;
        case HCE_WIFI_MSGID_SEND_IMAGE_START:                                        
        {
            DLNA_Printf(E_DEBUG, L_GENERAL,"Recevied MSG HCE_WIFI_MSGID_SEND_IMAGE_START \n");
            //reset status first
            Set_Upload_Status(UPLOAD_STATUS_IDLE);
            HCE_DLNA_MSGID_SEND_IMAGE_START_REQ_PARA* psSendImageStartInfo = 
                (HCE_DLNA_MSGID_SEND_IMAGE_START_REQ_PARA*) pvReqData;
            long size;
            FILE * pFile;
            int ret = 0;
            char converted_path[64];

            // convert file path from threadx to linux
            if( Convert_FilePath(converted_path, psSendImageStartInfo->content_full_path, OS_LINUX) < 0 ){
                psResponseInfo->eErrorResult = DLNA_REMOTE_MSG_RESULT_TYPE_GENERAL_ERROR;
                break;
            }
            
            upload_object_info.resize = psSendImageStartInfo->resize;
            upload_object_info.FilePathRoot = psSendImageStartInfo->FilePathRoot;
            
            if(psSendImageStartInfo->FilePathRoot == 1){ //sdcard
               strcpy(upload_object_info.content_full_path, DEF_SDCARD_ROOTPATH); 
            }else{ //disk c
               strcpy(upload_object_info.content_full_path, DEF_DISKC_ROOTPATH); 
            }

            strcat(upload_object_info.content_full_path, converted_path);
            if(psSendImageStartInfo->ContentBuffer != 0 && psSendImageStartInfo->resize)
                upload_object_info.ContentBuffer = DualosIpc_Mmap(psSendImageStartInfo->ContentBuffer);
            upload_object_info.nBufferSize = psSendImageStartInfo->nBufferSize;
            upload_object_info.nHeight = psSendImageStartInfo->nHeight;
            
            if(upload_object_info.resize != 1){
                pFile = fopen (upload_object_info.content_full_path,"rb");
                if (pFile==NULL) 
                    DLNA_Printf(E_DEBUG, L_GENERAL,"Opening file fail\n");
                else
                {
                    fseek (pFile, 0, SEEK_END);   // non-portable
                    size = ftell (pFile);
                    fclose (pFile);
                    DLNA_Printf(E_DEBUG, L_GENERAL,"Size of upload_object_info.content_full_path: %ld bytes.\n",size);
                }
                upload_object_info.nBufferSize = (uint32_t)size;
            }            upload_object_info.nWidth = psSendImageStartInfo->nWidth;
            
            DLNA_Printf(E_DEBUG, L_GENERAL,"path:%s %d %d %d \n",psSendImageStartInfo->content_full_path,psSendImageStartInfo->nBufferSize,psSendImageStartInfo->nWidth,psSendImageStartInfo->nHeight);     
            //Request_SOAP( "X_GetDLNAUploadProfiles" , "LUMIX_UP_SERVER_SMP" );
            //sleep(5);
            // only for test now

            Set_Upload_Status(UPLOAD_STATUS_TRANSFERRING);
            
            DLNA_Printf(E_DEBUG, L_GENERAL, "%lu, %lu, %s, %lu, %lu, %lu\n", upload_object_info.resize, upload_object_info.FilePathRoot, upload_object_info.content_full_path, \
                    upload_object_info.nBufferSize, upload_object_info.nWidth, upload_object_info.nHeight);
 		    ret = Request_SOAP( "CreateObject" , "LUMIX_UP_SERVER_SMP" );  
             
            if(ret != 0){
                DLNA_Printf(E_DEBUG, L_GENERAL,"Request_SOAP Fail\n");
                Set_Upload_Status(UPLOAD_STATUS_IDLE);
                psResponseInfo->eErrorResult = DLNA_REMOTE_MSG_RESULT_TYPE_GENERAL_ERROR;
            }
                
        }    
        break;
        case HCE_WIFI_MSGID_SEND_IMAGE_CANCEL:
        {
            DLNA_Printf(E_DEBUG, L_GENERAL,"Recevied MSG HCE_WIFI_MSGID_SEND_IMAGE_CANCEL \n");

            if(Get_Upload_Status() != UPLOAD_STATUS_IDLE){
                Set_Upload_Status(UPLOAD_STATUS_CANCELING);
                //wait status to be UPLOAD_STATUS_IDLE
                while(Get_Upload_Status() != UPLOAD_STATUS_IDLE){
                    DLNA_Printf(E_DEBUG, L_GENERAL,"wait UPLOAD STATUS TO IDLE \n");
                    usleep(100000); // wait 100 ms
                }
            }
            DLNA_Printf(E_DEBUG, L_GENERAL,"Leave MSG HCE_WIFI_MSGID_SEND_IMAGE_CANCEL \n");
            
        }    
        break;
        case HCE_WIFI_MSGID_SEND_IMAGE_PAUSE:
        {
            DLNA_Printf(E_DEBUG, L_GENERAL,"Recevied MSG HCE_WIFI_MSGID_SEND_IMAGE_PAUSE \n"); 
        }    
        break;
        case HCE_WIFI_MSGID_SEND_IMAGE_RESUME:
        {
            DLNA_Printf(E_DEBUG, L_GENERAL,"Recevied MSG HCE_WIFI_MSGID_SEND_IMAGE_RESUME \n"); 
        }    
        break;
        default :
        {
           DLNA_Printf(E_DEBUG, L_GENERAL,"WARNNING! Msg(0x%x) did not handle \n",psReqIPC->msgid);
        }    
        break;   
    }
    
}
//--------------------------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------------------------
//Send message and receive response.
//The content in request IPC content(sReqInfo.content) as below
//Response IPC content as the similar format.
//-----------------------------
//| DLNAIPCContentReqStruct   |
//-----------------------------
//|       Reqeust Data        |
//-----------------------------
int HceDLNA_IPC_SendMsg(DLNASendMessageStruct *psReqMsgData,DLNAResponseMessageStruct *psResMsgData)
{
    LNXIPC_GROUP_A_MSG_T     sReqIPC;
    DLNAIPCContentReqStruct* psReqInfo;
    memset(&sReqIPC,0,sizeof(LNXIPC_GROUP_A_MSG_T));
    LNXIPC_GROUP_A_MSG_T *psReqIPC = &sReqIPC;
    psReqInfo = (DLNAIPCContentReqStruct*) psReqIPC->content;

    
    LNXIPC_GROUP_A_MSG_T     sResponseIPC;
    DLNAIPCContentResStruct* psResponseInfo = NULL;
    memset(&sResponseIPC,0,sizeof(LNXIPC_GROUP_A_MSG_T));
    LNXIPC_GROUP_A_MSG_T *psResponseIPC = &sResponseIPC;
    psResponseInfo = (DLNAIPCContentResStruct*) psResponseIPC->content;

    //Error handle
    if(psReqMsgData->ulReqParaLen > (MAIN_QUEUE_GROUP_A_MAX_CONTENT_SIZE - sizeof(DLNAIPCContentReqStruct))) {
        DLNA_Printf(E_ERROR, L_GENERAL, "ERROR!! Input(%d) is more than %d\n",psReqMsgData->ulReqParaLen,MAIN_QUEUE_GROUP_A_MAX_CONTENT_SIZE);
        return -1;
    }
    
    if(psResMsgData->ulResParaLen > (MAIN_QUEUE_GROUP_A_MAX_CONTENT_SIZE - sizeof(DLNAIPCContentResStruct)))  {
        DLNA_Printf(E_ERROR, L_GENERAL, "ERROR!! Output(%d) is more than %d\n",psResMsgData->ulResParaLen,MAIN_QUEUE_GROUP_A_MAX_CONTENT_SIZE);
        return -1;
    }

    //construct request data
    psReqIPC->msgid               = (uint32_t)psReqMsgData->eMsgId;
    psReqInfo->ulReqDataSize      = psReqMsgData->ulReqParaLen;
    psReqInfo->ulResponseDataSize = psResMsgData->ulResParaLen;

    if(psReqMsgData->pvReqPara && psReqMsgData->ulReqParaLen)
        memcpy((void*)&psReqIPC->content[sizeof(DLNAIPCContentReqStruct)],(void*)psReqMsgData->pvReqPara,psReqMsgData->ulReqParaLen);

        
    LnxIpcGroupA_TunnelSend(pDLNAWiFiInterfaceInfo->pvIPCFD, GROUP_A_SERVERID_WIFI_FOR_DLNA, eCLIENT_SEND_MODE_BLOCK, 0, psReqIPC, (void*)&psResponseIPC);

    //construct response data
    if(psResMsgData->pvResPara && psResMsgData->ulResParaLen)
        memcpy((void*)psResMsgData->pvResPara,(void*)&psResponseIPC->content[sizeof(DLNAIPCContentResStruct)],psResMsgData->ulResParaLen);

    DLNA_Printf(E_DEBUG, L_GENERAL, "the req size:%d, SS:%d, res size :%d SS:%d RS:%d\n",
                psReqInfo->ulReqDataSize,
                sizeof(DLNAIPCContentReqStruct),
                psReqInfo->ulResponseDataSize,
                sizeof(DLNAIPCContentResStruct),
                psResponseInfo->eErrorResult
                );
    return 0;
}

//-----------------------------------------------------------------------------------------------------
//Handle message if necessary
int HceDLNA_Remote_SendMsg(DLNASendMessageStruct *psRequeestStruct,DLNAResponseMessageStruct *psResponseStruct)
{
  int lExecRsult    = 0;
  switch(psRequeestStruct->eMsgId)
  {
    case HCE_WIFI_MSGID_CGI_PLAYCMD_DELETE_FILE:
        DLNA_Printf(E_INFO, L_GENERAL, "HCE_WIFI_MSGID_CGI_PLAYCMD_DELETE_FILE\n");
    break;
    case HCE_WIFI_MESID_CGI_PLAYCMD_GET_GROUP_INFROMATION:  
        DLNA_Printf(E_INFO, L_GENERAL, "HCE_WIFI_MESID_CGI_PLAYCMD_GET_GROUP_INFROMATION\n");
    break;
    default :        
        DLNA_Printf(E_INFO, L_GENERAL, "Get Message 0x%x\n",psRequeestStruct->eMsgId);
    break;    
  }
  
  lExecRsult = HceDLNA_IPC_SendMsg(psRequeestStruct,psResponseStruct);

  DLNA_Printf(E_INFO, L_GENERAL, "Message 0x%x is sent\n",psRequeestStruct->eMsgId);
  
  if(lExecRsult <0)  {
      DLNA_ErrorPrintf("WARNING!! fail to send message\n");
  }  
  return lExecRsult;
}

//-----------------------------------------------------------------------------------------------------

int HceDLNA_GeneralMessageToWiFi(Hce_WiFi_MsgID_Type eMssageID,void* pvReqPtr,unsigned int ulReqLen,void* pvResPtr,unsigned int ulResLen)
{
   int lResult = 0;
   DLNASendMessageStruct     sRequestData;
   DLNAResponseMessageStruct sResponseData;

   memset(&sRequestData ,0,sizeof(DLNASendMessageStruct));
   memset(&sResponseData,0,sizeof(DLNAResponseMessageStruct));

   sRequestData.eMsgId        = eMssageID;
   sRequestData.pvReqPara     = pvReqPtr;
   sRequestData.ulReqParaLen  = ulReqLen;

   sResponseData.pvResPara    = pvResPtr;
   sResponseData.ulResParaLen = ulResLen;

   lResult = HceDLNA_Remote_SendMsg(&sRequestData,&sResponseData);
   return lResult;

}


//-----------------------------------------------------------------------------------------------------

int HceDLNA_IPC_Init(void)
{
    pDLNAWiFiInterfaceInfo = malloc(sizeof(DLNAWiFiInterfaceInfoStruct));
    memset(pDLNAWiFiInterfaceInfo,0,sizeof(DLNAWiFiInterfaceInfoStruct));
    
    if(pDLNAWiFiInterfaceInfo == NULL)  {
        DLNA_Printf(E_ERROR, L_GENERAL,"ERROR fail to create buffer\n");
        return -1;
    }    

    pDLNAWiFiInterfaceInfo->pvIPCFD = LnxIpcGroupA_TunnelInit(GROUP_A_SERVERID_DLNA,LNXIPC_THREAD_PRIORITY_NORMAL,HceDLNA_IPCServerReceiveCB);

    if(pDLNAWiFiInterfaceInfo->pvIPCFD == NULL)  {    
        DLNA_Printf(E_ERROR, L_GENERAL,"ERROR!! Fail to create DLNA IPC\n");
        return -1;
    }    
    
    DLNA_Printf(E_DEBUG, L_GENERAL,"IPC init done\n");
    return 0;
}    

int HceDLNA_IPC_DeInit(void)
{
    if(pDLNAWiFiInterfaceInfo == NULL)  {   
        DLNA_Printf(E_WARN, L_GENERAL," WARNING buffer is NULL\n"); 
    }
    else  {        
        free(pDLNAWiFiInterfaceInfo);
        pDLNAWiFiInterfaceInfo = NULL;
    }    

    if(pDLNAWiFiInterfaceInfo->pvIPCFD == NULL)  {    
        DLNA_Printf(E_WARN, L_GENERAL," WARNING IPC did not initialized\n"); 
    }
    else  {        
        LnxIpcGroupA_TunnelDeinit(pDLNAWiFiInterfaceInfo->pvIPCFD);
        pDLNAWiFiInterfaceInfo->pvIPCFD = NULL;
    }

  
    DLNA_Printf(E_DEBUG, L_GENERAL,"IPC DE-init done\n");
    return 0;
}    


//-----------------------------------------------------------------------------------------------------